@@ -26,16 +26,17 @@ import subprocess
 import sys
 import time
 from collections import OrderedDict
-from pathlib import Path
 from typing import Dict
 
 import requests
 
 from db_contrib_tool.plugin import PluginInterface, Subcommand, SubcommandResult
-
-sys.path.append(str(Path(os.getcwd(), __file__).parent.parent))
 from db_contrib_tool.utils.build_system_options import PathOptions
-from db_contrib_tool.utils.oauth import Configs, get_oauth_credentials
+from db_contrib_tool.utils.oauth import (
+    Configs,
+    get_client_cred_oauth_credentials,
+    get_oauth_credentials,
+)
 
 SUBCOMMAND = "symbolize"
 
@@ -124,6 +125,7 @@ class PathResolver(object):
         client_credentials_scope: str = None,
         client_credentials_user_name: str = None,
         client_id: str = None,
+        client_secret: str = None,
         redirect_port: int = None,
         scope: str = None,
         auth_domain: str = None,
@@ -146,6 +148,7 @@ class PathResolver(object):
             client_credentials_user_name or self.default_client_credentials_user_name
         )
         self.client_id = client_id
+        self.client_secret = client_secret
         self.redirect_port = redirect_port
         self.scope = scope
         self.auth_domain = auth_domain
@@ -175,11 +178,19 @@ class PathResolver(object):
                 data = json.loads(cfile.read())
                 access_token, expire_time = data.get("access_token"), data.get("expire_time")
                 if time.time() < expire_time:
-                    # credentials hasn't expired yet
+                    # credentials not expired yet
                     self.http_client.headers.update({"Authorization": f"Bearer {access_token}"})
                     return
 
-        credentials = get_oauth_credentials(configs=self.configs, print_auth_url=True)
+        if self.client_id and self.client_secret:
+            # auth using secrets
+            credentials = get_client_cred_oauth_credentials(
+                self.client_id, self.client_secret, self.configs
+            )
+        else:
+            # since we don't have access to secrets, ask user to auth manually
+            credentials = get_oauth_credentials(configs=self.configs, print_auth_url=True)
+
         self.http_client.headers.update({"Authorization": f"Bearer {credentials.access_token}"})
 
         # write credentials to local file for further useage
@@ -262,7 +273,11 @@ class PathResolver(object):
         filename = self.url_to_filename(url)
         path = os.path.join(self.cache_dir, filename)
         if not os.path.exists(path):
-            subprocess.check_call(["wget", url], cwd=self.cache_dir)
+            print("Downloading the file...")
+            with requests.get(url, stream=True) as response:
+                with open(path, "wb") as file:
+                    for chunk in response.iter_content(chunk_size=2 * 1024 * 1024):
+                        file.write(chunk)
         else:
             print("File aready exists in cache")
             exists_locally = True
@@ -276,15 +291,17 @@ class PathResolver(object):
         :return: path as string or None (if path not found)
         """
         build_id = soinfo.get("buildId", "").lower()
+        version = soinfo.get("version")
         binary_name = "mongo"
         # search from cached results
         path = self.get_from_cache(build_id)
         if not path:
             # path does not exist in cache, so we send request to server
             try:
-                response = self.http_client.get(
-                    f"{self.host}/find_by_id", params={"build_id": build_id}
-                )
+                search_parameters = {"build_id": build_id}
+                if version:
+                    search_parameters["version"] = version
+                response = self.http_client.get(f"{self.host}/find_by_id", params=search_parameters)
                 if response.status_code != 200:
                     # if we could not find the path of binary, that might be system library.
                     # we can try using frame's own `path` data.
@@ -325,7 +342,7 @@ class PathResolver(object):
             if exists_locally:
                 path = dl_path.replace(".tgz", "", 1)
             else:
-                print("Downloaded, now unpacking...")
+                sys.stdout.write("Downloaded, now unpacking...\n")
                 path = self.unpack(dl_path)
         except Exception as err:  # noqa
             sys.stderr.write(f"Failed to download & unpack file: {err}\n")
@@ -333,7 +350,7 @@ class PathResolver(object):
         # if file has extension, it is good. if not, we should append .debug, because those without extension are
         # from release builds, and their debug symbol files contain .debug extension.
         # we need to map those 2 different file names ('<name>' becomes '<name>.debug').
-        if not binary_name.endswith(".debug") and not binary_name.endswith(".so"):
+        if not binary_name.endswith(".debug"):
             binary_name = f"{binary_name}.debug"
 
         inner_folder_name = self.path_options.get_binary_folder_name(binary_name)
@@ -352,6 +369,8 @@ class Symbolizer(Subcommand):
         dsym_hint,
         symbolizer_path,
         live,
+        client_secret,
+        client_id,
     ):
         self.output_format = output_format
         self.host = host
@@ -361,6 +380,8 @@ class Symbolizer(Subcommand):
         self.dsym_hint = dsym_hint
         self.symbolizer_path = symbolizer_path
         self.live = live
+        self.client_secret = client_secret
+        self.client_id = client_id
 
     def substitute_stdin(self, resolver):
         """Accept stdin stream as source of logs and symbolize it."""
@@ -394,11 +415,17 @@ class Symbolizer(Subcommand):
                 print(prefix)
                 print("Symbolizing...")
                 self.classic_output(frames, sys.stdout, indent=2)
+                print("Symbolization completed, waiting for input...")
             else:
                 print(line)
 
     def execute(self):
-        resolver = PathResolver(host=self.host, cache_dir=self.cache_dir)
+        resolver = PathResolver(
+            host=self.host,
+            cache_dir=self.cache_dir,
+            client_secret=self.client_secret,
+            client_id=self.client_id,
+        )
 
         if self.live:
             print("Entering live mode")
@@ -486,6 +513,9 @@ class Symbolizer(Subcommand):
             return {so_entry["b"]: so_entry for so_entry in somap_list if "b" in so_entry}
 
         base_addr_map = make_base_addr_map(trace_doc["processInfo"]["somap"])
+        version = trace_doc.get("processInfo", {}).get("mongodbVersion")
+        if version and "patch" in version:
+            version = version.split("-")[-1]
 
         frames = []
         for frame in trace_doc["backtrace"]:
@@ -495,6 +525,8 @@ class Symbolizer(Subcommand):
                 )
                 continue
             soinfo = base_addr_map.get(frame["b"], {})
+            if version:
+                soinfo["version"] = version
             elf_type = soinfo.get("elfType", 0)
             if elf_type == 3:
                 addr_base = "0"
@@ -504,7 +536,7 @@ class Symbolizer(Subcommand):
                 addr_base = soinfo.get("vmaddr", "0")
             addr = int(addr_base, 16) + int(frame["o"], 16)
             # addr currently points to the return address which is the one *after* the call. x86 is
-            # variable length so going backwards is difficult. However llvm-symbolizer seems to do the
+            # variable length so going backwards is difficult. However, llvm-symbolizer seems to do the
             # right thing if we just subtract 1 byte here. This has the downside of also adjusting the
             # address of instructions that cause signals (such as segfaults and divide-by-zero) which
             # are already correct, but there doesn't seem to be a reliable way to detect that case.
@@ -528,11 +560,17 @@ class Symbolizer(Subcommand):
         # Keep frames in kwargs to avoid changing the function signature.
         frames = kwargs.get("frames")
         if frames is None:
-            frames = self.preprocess_frames(dbg_path_resolver, trace_doc, input_format)
+            total_seconds_for_retries = kwargs.get("total_seconds_for_retries", 0)
+            frames = self.preprocess_frames(
+                dbg_path_resolver,
+                trace_doc,
+                input_format,
+                total_seconds_for_retries=total_seconds_for_retries,
+            )
 
         if not symbolizer_path:
             symbolizer_path_env = "MONGOSYMB_SYMBOLIZER_PATH"
-            default_symbolizer_path = "llvm-symbolizer"
+            default_symbolizer_path = "/opt/mongodbtoolchain/v4/bin/llvm-symbolizer"
             symbolizer_path = os.environ.get(symbolizer_path_env)
             if not symbolizer_path:
                 print(
@@ -549,7 +587,7 @@ class Symbolizer(Subcommand):
             close_fds=True,
             stdin=subprocess.PIPE,
             stdout=subprocess.PIPE,
-            stderr=open("/dev/null"),
+            stderr=sys.stdout,
         )
 
         def extract_symbols(stdin):
@@ -592,8 +630,23 @@ class Symbolizer(Subcommand):
         symbolizer_process.wait()
         return frames
 
-    def preprocess_frames(self, dbg_path_resolver, trace_doc, input_format):
-        """Process the paths in frame objects."""
+    def preprocess_frames(
+        self,
+        dbg_path_resolver,
+        trace_doc,
+        input_format,
+        total_seconds_for_retries=0,
+        current_retry_number=0,
+    ):
+        """
+        Process the paths in frame objects.
+
+        :param dbg_path_resolver: resolver class
+        :param trace_doc: traceback string
+        :param input_format: format of input
+        :param total_seconds_for_retries: max wait time for retries in seconds
+        :param current_retry_number: current retry counter
+        """
         if input_format == "classic":
             frames = self.parse_input(trace_doc, dbg_path_resolver)
         elif input_format == "thin":
@@ -602,6 +655,61 @@ class Symbolizer(Subcommand):
                 frame["path"] = dbg_path_resolver.get_dbg_file(frame)
         else:
             raise ValueError('Unknown input format "{}"'.format(input_format))
+
+        # Check if path resolver actually found enough paths.
+        # If (not found / all) ratio is higher than 0.5, we retry.
+        not_found = [1 for i in frames if i.get("path") is None]
+        not_found_ratio = len(not_found) / (len(frames) or 1)
+        # When we fail to find more than 50% of paths (not_found_ratio >= 0.5),
+        # we wait for specific amount of time,
+        # and then retry again. this continues until we hit total_seconds_for_retries.
+        # After each wait, we increment next wait time, to not flood with too many retries with the same interval.
+        # There was thought about incrementing exponentially, by multiplying by 2 (time *= 2),
+        # but that step seemed too high, making less breakpoints to stop and retry.
+        # Another idea was to add 1 minute each time, but that felt like too frequent.
+        # The final decision was to go with 2: 2 extra minutes after each wait.
+        increment_mins = 2
+
+        if not_found_ratio >= 0.5 and total_seconds_for_retries > 0:
+            # Retry processing.
+            # Our total wait time should not exceed total_seconds_for_retries.
+            # We should calculate how long we already waited up to this point.
+            # Since we increment wait time by 2 minutes each time,
+            # it makes a perfect arithmetic sequence.
+            # We can calculate the sum of sequence using this formula: S = n/2 (a + L)
+            # where: S - sum, n - number of items in sequence, a - first term, L - last term.
+            # Items in our sequence is equal to current_retry_number.
+            # First term is always 2, which is the increment_mins value.
+            # Last term is current_retry_number.
+            # Formula: S = current_retry_number/2 (2 + current_retry_number).
+            # Do not forget to multiply by 60 to convert into seconds.
+            total_wait_time = (
+                current_retry_number / 2 * (increment_mins + current_retry_number) * 60
+            )
+            if total_wait_time > total_seconds_for_retries:
+                # We already waited/tried enough.
+                return frames
+
+            current_retry_number += 1
+            # We can multiply current_retry_number by increment_mins value,
+            # and it will be the same as current_retry_number += increment_mins
+            wait_time = current_retry_number * increment_mins * 60
+            if (total_wait_time + wait_time) > total_seconds_for_retries:
+                # If we wait for another wait_time, our total wait time will exceed total_seconds_for_retries.
+                # We don't want that. so, we justify it to the amount where it doesn't exceed the limit.
+                wait_time = total_seconds_for_retries - total_wait_time
+            print(
+                f"Currently {not_found_ratio * 100:.1f}% of frames couldn't be resolved. "
+                f"Retrying again in {wait_time} seconds..."
+            )
+            time.sleep(wait_time)
+            return self.preprocess_frames(
+                dbg_path_resolver,
+                trace_doc,
+                input_format,
+                total_seconds_for_retries,
+                current_retry_number,
+            )
         return frames
 
     @staticmethod
@@ -638,6 +746,8 @@ class SymbolizerPlugin(PluginInterface):
             dsym_hint=parsed_args.dsym_hint,
             symbolizer_path=parsed_args.symbolizer_path,
             live=parsed_args.live,
+            client_secret=parsed_args.client_secret,
+            client_id=parsed_args.client_id,
         )
 
     def add_subcommand(self, subparsers):
@@ -659,6 +769,16 @@ class SymbolizerPlugin(PluginInterface):
             help="Specify a src dir to move to /data/mci/{original_buildid}/src",
         )
 
+        parser.add_argument(
+            "--total-seconds-for-retries",
+            default=0,
+            type=int,
+            help="If web service fails to find path for given build id, it could be because mapping "
+            "process was not finished yet. We can wait for it to finish and retry again. Each retry"
+            " adds 2 minutes to previous wait time. It is guaranteed that total wait time does not exceed this "
+            "specified amount.",
+        )
+
         parser.add_argument("--live", action="store_true")
         parser.add_argument(
             "--host", default="", help="URL of web service running the API to get debug symbol URL"
@@ -666,6 +786,8 @@ class SymbolizerPlugin(PluginInterface):
         parser.add_argument(
             "--cache-dir", default="", help="Full path to a directory to store cache/files"
         )
+        parser.add_argument("--client-secret", default="", help="Secret key for Okta Oauth")
+        parser.add_argument("--client-id", default="", help="Client id for Okta Oauth")
         # caching mechanism is currently not fully developed and needs more advanced cleaning techniques,
         # we add an option to enable it after completing the implementation.
 
